using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Infragistics.Win.UltraWinGrid;
using System.Threading;

namespace Bubblebreaker
{
    /// <summary>
    ///     Bubble breaker.
    /// </summary>
    public partial class UCBubbleBreaker : UserControl
    {
        #region Const
        /// <summary>
        ///        .
        /// </summary>
        private static readonly int COLUMN_WIDTH = 22;

        /// <summary>
        ///     5 . 
        /// </summary>
        private static readonly int TOTAL_COLORS = 5;

        #endregion

        #region 
        /// <summary>
        ///  .
        /// </summary>
        private Matrix field;

        /// <summary>
        ///   .
        /// </summary>
        private int totalScore;

        /// <summary>
        ///      .
        /// </summary>
        private string fileToSaveStatistics;
        #endregion

        #region 
        /// <summary>
        ///  ,      .
        /// </summary>
        private int CurrentTurnScore
        {
            set
            {
                lCuurentTurbScore.Text = "    : " + value;
            }
        }

        /// <summary>
        ///   .
        /// </summary>
        private int TotalScore
        {
            get { return totalScore; }
            set
            {
                totalScore = value;
                lTotalCount.Text = string.Format(": {0}", totalScore);
            }
        }

        /// <summary>
        ///      .
        /// </summary>
        public string FileToSaveStatistics
        {
            get { return fileToSaveStatistics; }
            set { fileToSaveStatistics = value; }
        }
        #endregion

        #region   
        /// <summary>
        /// ,    .
        /// </summary>
        /// <param name="totalScore"> .</param>
        public delegate void GameOverDelegate(int totalScore);

        /// <summary>
        ///     .
        /// </summary>
        public event GameOverDelegate OnGameOver;
        #endregion

        #region 
        /// <summary>
        /// 
        /// </summary>
        public UCBubbleBreaker()
        {
            InitializeComponent();
            field = new Matrix(0, 0);
            field.OnUpdated += new EventHandler(field_OnUpdated);
            grid.DataSource = field.Table;
            grid.DisplayLayout.Bands[0].ColHeadersVisible = false;
        }
        #endregion

        #region 
        /// <summary>
        ///  .
        /// </summary>
        public void Reset()
        {
            grid.DataSource = field.Table;
            foreach (UltraGridColumn column in grid.DisplayLayout.Bands[0].Columns)
            {
                column.Width = COLUMN_WIDTH;
                column.ValueList = grid.DisplayLayout.ValueLists[0];
            }
            CurrentTurnScore = 0;
            TotalScore = 0;
        }


        /// <summary>
        ///  .
        /// </summary>
        /// <param name="columns"> .</param>
        /// <param name="rows"> .</param>
        public void Create(int columns, int rows)
        {
            field = new Matrix(columns, rows);
            field.InitMatrixWithNumber(TOTAL_COLORS);
            Reset();
        }

        #region field.    .
        /// <summary>
        ///    .
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void field_OnUpdated(object sender, EventArgs e)
        {
            grid.Selected.Cells.Clear();
            grid.Refresh();
        }
        #endregion

        #region grid.  .
        /// <summary>
        ///  double click    .
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_DoubleClickCell(object sender, Infragistics.Win.UltraWinGrid.DoubleClickCellEventArgs e)
        {
            this.BeginInvoke(new MethodInvoker(delegate
            {
                TotalScore += field.MakeTurn(e.Cell.Row.Index, e.Cell.Column.Index);
                grid.Selected.Cells.Clear();
                if (field.IsGameOver && OnGameOver != null)
                {
                    field.Clear();
                    OnGameOver(totalScore);
                }
            }));
        }

        /// <summary>
        ///        ,     .
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_BeforeCellActivate(object sender, CancelableCellEventArgs e)
        {
            grid.Selected.Cells.Clear();
            List<KeyValuePair<int, int>> cellsToSelect = field.GetCellsForTurn(e.Cell.Row.Index, e.Cell.Column.Index);
            CurrentTurnScore = field.ScoreByCount(cellsToSelect.Count);
            cellsToSelect.ForEach(delegate(KeyValuePair<int, int> x)
            {
                grid.Selected.Cells.Add(grid.Rows[x.Key].Cells[x.Value]);
            });
            e.Cancel = true;
        }

        /// <summary>
        ///    tooltip'.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_InitializeLayout(object sender, InitializeLayoutEventArgs e)
        {
            e.Layout.Override.TipStyleCell = TipStyle.Hide;
        }
        #endregion

        #endregion
    }


    #region Class. Matrix -    

    /// <summary>
    ///        Bubble breaker.
    /// </summary>
    class Matrix
    {
        #region Const
        /// <summary>
        ///  ()    ,  " ",         .
        /// </summary>
        private static readonly int TRY_COLOR = -1;

        /// <summary>
        ///  ()  .
        /// </summary>
        private static readonly int EMPTY_COLOR = 0;

        #endregion

        public event EventHandler OnUpdated;

        /// <summary>
        /// 
        /// </summary>
        private DataTable table;

        #region 

        /// <summary>
        ///    DataTable  binding  .
        /// </summary>
        public DataTable Table
        {
            get { return table; }
        }

        /// <summary>
        /// ,   .
        /// </summary>
        /// <returns>true,     .</returns>
        public bool IsGameOver
        {
            get
            {
                for (int i = 0; i < rows; i++)
                {
                    for (int j = 0; j < columns; j++)
                    {
                        if (this[i, j] == EMPTY_COLOR)
                            continue;

                        if (GetCellsForTurn(i, j).Count > 1)
                            return false;
                    }
                }
                return true;
            }
        }

        /// <summary>
        ///   .
        /// </summary>
        public int rows
        {
            get { return table.Rows.Count; }
        }

        /// <summary>
        ///   .
        /// </summary>
        public int columns
        {
            get { return table.Columns.Count; }
        }

        /// <summary>
        ///     
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public int this[int x, int y]
        {
            get { return (int)table.Rows[x][y]; }
            set { table.Rows[x][y] = value; }
        }

        #endregion

        #region 
        /// <summary>
        /// .
        /// </summary>
        /// <param name="columns"></param>
        /// <param name="rows"></param>
        public Matrix(int columns, int rows)
        {
            this.table = new DataTable();
            for (int i = 0; i < columns; i++)
                table.Columns.Add(new DataColumn(i.ToString(), typeof(int)));
            for (int j = 0; j < rows; j++)
                table.Rows.Add();
        } 
        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="colorNumber"></param>
        public void InitMatrixWithNumber(int colorNumber)
        {
            Random rnd = new Random(DateTime.Now.Millisecond);
            for (int i = 0; i < rows; i++)
                for (int j = 0; j < columns; j++)
                    this[i, j] = rnd.Next(colorNumber) + 1;  // +1     
        }


        /// <summary>
        ///   ,     .     .
        /// </summary>
        /// <param name="x">X   .</param>
        /// <param name="y">Y   .</param>
        /// <returns>   .</returns>
        public List<KeyValuePair<int, int>> GetCellsForTurn(int x, int y)
        {
            int tmpColor = this[x, y];
            if (tmpColor == EMPTY_COLOR)
                return new List<KeyValuePair<int, int>>();
            FloodFill(x, y, TRY_COLOR);
            List<KeyValuePair<int, int>> ret = GetCellsByValue(TRY_COLOR);
            FloodFill(x, y, tmpColor);
            return ret;
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private List<KeyValuePair<int, int>> GetCellsByValue(int value)
        {
            List<KeyValuePair<int, int>> ret = new List<KeyValuePair<int, int>>();
            for (int i = 0; i < rows; i++)
                for (int j = 0; j < columns; j++)
                    if (this[i, j] == value)
                        ret.Add(new KeyValuePair<int, int>(i, j));
            return ret;
        }

        /// <summary>
        ///     floodfill.
        /// </summary>
        /// <param name="x">X   .</param>
        /// <param name="y">Y   .</param>
        /// <param name="targetValue">  .</param>
        /// <param name="replacementValue">   .</param>
        /// <param name="count">   .</param>
        private void FloodFillStep(int x, int y, int targetValue, int replacementValue, ref int count)
        {
            if (x < 0 || x > rows - 1 || y < 0 || y > columns - 1 || this[x, y] != targetValue || this[x, y] == replacementValue)
                return;

            this[x, y] = replacementValue;
            count++;

            FloodFillStep(x - 1, y, targetValue, replacementValue, ref count);
            FloodFillStep(x + 1, y, targetValue, replacementValue, ref count);
            FloodFillStep(x, y - 1, targetValue, replacementValue, ref count);
            FloodFillStep(x, y + 1, targetValue, replacementValue, ref count);
        }

        /// <summary>
        ///      flood fill.
        /// </summary>
        /// <param name="x">X   .</param>
        /// <param name="y">Y   .</param>
        /// <param name="replacementValue">  .</param>
        /// <returns>   .</returns>
        private int FloodFill(int x, int y, int replacementValue)
        {
            int count = 0;
            FloodFillStep(x, y, this[x, y], replacementValue, ref count);
            return count;
        }


        /// <summary>
        ///       .
        /// </summary>
        /// <param name="x">X   .</param>
        /// <param name="y">Y   .</param>
        /// <returns>      :)</returns>
        public int MakeTurn(int x, int y)
        {
            //      
            if (GetCellsForTurn(x, y).Count <= 1)
                return 0;
            else
            {
                int ret = ScoreByCount(FloodFill(x, y, EMPTY_COLOR));
                UpdateField();
                Thread.Sleep(300);
                PullDownAllColumns();
                UpdateField();
                Thread.Sleep(300);
                MoveEmptyColumns();
                UpdateField();
                return ret;
            }
        }


        /// <summary>
        /// 
        /// </summary>
        private void UpdateField()
        {
            if (OnUpdated != null)
                OnUpdated(this, null);
        }

        /// <summary>
        ///   ,       n*(n-1).
        /// </summary>
        /// <param name="processedCellcCount">  .</param>
        /// <returns></returns>
        public int ScoreByCount(int processedCellcCount)
        {
            return processedCellcCount * (processedCellcCount - 1);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder((rows + 1) * 2 * columns + 10);
            for (int i = 0; i < rows; i++)
            {
                for (int j = 0; j < columns; j++)
                {
                    sb.Append(this[i, j]);
                    sb.Append(' ');
                }
                sb.Append(Environment.NewLine);
            }
            return sb.ToString();
        }


        /// <summary>
        ///      ,  ""  .
        /// </summary>
        private void PullDownAllColumns()
        {
            for (int i = 0; i < columns; i++)
            {
                for (int j = rows - 1; j >= 0; j--)
                {
                    if (this[j, i] != EMPTY_COLOR)
                        continue;

                    //  ,          ""  .
                    int pulldownCell = -1;
                    for (int k = j; k >= 0; k--)
                    {
                        if (this[k, i] != EMPTY_COLOR)
                        {
                            pulldownCell = k;
                            break;
                        }
                    }

                    //    ,     .
                    if (pulldownCell == -1)
                        break;

                    int tmp = this[j, i];
                    this[j, i] = this[pulldownCell, i];
                    this[pulldownCell, i] = tmp;
                }
            }
        }



        /// <summary>
        ///     .
        /// </summary>
        private void MoveEmptyColumns()
        {
            for (int i = columns - 1; i >= 0; i--)
            {
                if (IsEmptyColumn(i))
                    MoveAllColumnsRight(i);
            }
        }


        /// <summary>
        ///       .
        /// </summary>
        /// <param name="column"> .</param>
        /// <returns>true,    </returns>
        private bool IsEmptyColumn(int column)
        {
            for (int j = rows - 1; j >= 0; j--)
                if (this[j, column] != EMPTY_COLOR)
                    return false;
            return true;
        }



        /// <summary>
        ///    ,   .
        /// </summary>
        /// <param name="column">     .</param>
        private void MoveAllColumnsRight(int column)
        {
            for (int i = 0; i < rows; i++)
            {
                for (int j = column; j > 0; j--)
                    this[i, j] = this[i, j - 1];

                this[i, 0] = EMPTY_COLOR;
            }
        }


        /// <summary>
        ///  .
        /// </summary>
        internal void Clear()
        {
            for (int i = 0; i < rows; i++)
                for (int j = 0; j < columns; j++)
                    this[i, j] = EMPTY_COLOR;
        }
    }
    #endregion

    #region Class. GameStatistics -     . 
    /// <summary>
    ///     .
    /// </summary>
    class GameStatistics
    {
        /// <summary>
        /// .
        /// </summary>
        KeyValuePair<string, UserGameResults>[] Statistics;

        /// <summary>
        ///    .
        /// </summary>
        private string fileStatisticsPath;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="fileStatisticsPath"></param>
        public GameStatistics(string fileStatisticsPath)
        {
            LoadFromFile(fileStatisticsPath);

        }

        /// <summary>
        ///    .
        /// </summary>
        /// <param name="user"></param>
        /// <param name="score"></param>
        public void Add(string user, int score)
        {
            bool exist = false;
            for (int i = 0; i < Statistics.Length; i++)
            {
                if (Statistics[i].Key.Equals(user, StringComparison.InvariantCultureIgnoreCase))
                {
                    Statistics[i].Value.AddResult(score);
                    exist = true;
                }
            }
            if (!exist)
            {
                List<KeyValuePair<string, UserGameResults>> tmp = (new List<KeyValuePair<string, UserGameResults>>(Statistics));
                tmp.Add(new KeyValuePair<string, UserGameResults>(user, new UserGameResults(user, new GameResult(DateTime.Now, score))));
                Statistics = tmp.ToArray();
            }

            SaveToFile(fileStatisticsPath);
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="filePath"></param>
        public void SaveToFile(string filePath)
        {

        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="filePath"></param>
        public void LoadFromFile(string filePath)
        {

        }
    } 
    #endregion

    #region Struct. GameResult - ,   .
    /// <summary>
    /// ,   .
    /// </summary>
    struct GameResult
    {
        /// <summary>
        ///   .
        /// </summary>
        DateTime Date;

        /// <summary>
        /// .
        /// </summary>
        int Score;

        /// <summary>
        /// .
        /// </summary>
        /// <param name="date">  .</param>
        /// <param name="score">.</param>
        public GameResult(DateTime date, int score)
        {
            Score = score;
            Date = date;
        }
    }
    #endregion

    #region Class. UserGameResults - ,     . 
    /// <summary>
    ///      .
    /// </summary>
    class UserGameResults
    {
        /// <summary>
        ///   .
        /// </summary>
        private static readonly int HISTORY_LENGTH = 5;

        /// <summary>
        ///      .
        /// </summary>
        public string UserName;

        /// <summary>
        ///     ;
        /// </summary>
        public int MaxResult;

        /// <summary>
        ///   .
        /// </summary>
        public GameResult[] Results;

        /// <summary>
        ///       .
        /// </summary>
        /// <param name="UserName"> .</param>
        /// <param name="Results">  .</param>
        public UserGameResults(string UserName, GameResult[] Results)
        {
            this.UserName = UserName;
            this.Results = Results;
        }

        /// <summary>
        ///     .
        /// </summary>
        /// <param name="UserName"> .</param>
        /// <param name="firstResult">   .</param>
        public UserGameResults(string UserName, GameResult firstResult)
        {
            this.UserName = UserName;
            this.Results = new GameResult[] { firstResult };
        }

        /// <summary>
        ///    .
        /// </summary>
        /// <param name="score">  .</param>
        public void AddResult(int score)
        {
            List<GameResult> tmp = new List<GameResult>(Results);
            tmp.Insert(0, new GameResult(DateTime.Now, score));
            Results = tmp.ToArray();
            if (Results.Length >= HISTORY_LENGTH)
                Array.Resize<GameResult>(ref Results, HISTORY_LENGTH);
            if (score > MaxResult)
                MaxResult = score;
        }
    } 
    #endregion

}
